/* 
 * $Id: route_atob.c,v 1.4 1998/05/11 17:32:46 masaki Exp $
 */

#include <version.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <netdb.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>
#include <mrt.h>
#include <array.h>
#include <hash.h>
#include <lbgp.h>
#include <bgp.h>
#include <io.h>

static void read_ascii (char *name, io_t * io);
static void skip (FILE * fd);

#ifndef HAVE_STRPTIME
char *strptime (char *buf, const char *format, struct tm *tm);
#endif /* HAVE_STRPTIME */


void
main (int argc, char *argv[])
{
    char c;
    extern char *optarg;	/* getopt stuff */
    extern int optind;		/* getopt stuff */
    char *usage =
    "Usage: route_atob [-i ascii_data_file] [-(o|w) binary_out_file]\n";
    char *filename = "stdin";
    int errors = 0;
    io_t *IO;
    trace_t *default_trace;

    default_trace = New_Trace ();
    IO = New_IO (default_trace);
    io_set (IO, IO_OUTFILE, "stdout", NULL);

    while ((c = getopt (argc, argv, "i:f:o:w:v")) != -1)
	switch (c) {
	case 'o':
	    if (io_set (IO, IO_OUTFILE, (char *) (optarg), NULL) < 0) {
		printf ("Error setting outfile\n");
		errors++;
	    }
	    break;
	case 'w':
	    if (io_set (IO, IO_OUTMSGQ, (char *) (optarg), NULL) < 0) {
		printf ("Error setting output %s\n", optarg);
		errors++;
	    }
	    break;
	case 'i':
	case 'f':
	    filename = (char *) optarg;
	    break;
	case 'v':
	    set_trace (default_trace, TRACE_FLAGS, TR_ALL,
		       TRACE_LOGFILE, "stdout", NULL);
	    break;
	default:
	    errors++;
	}

    if (errors) {
	fprintf (stderr, usage);
	printf ("\nMRT version (%s) compiled on %s\n\n",
		MRT_VERSION, __DATE__);
	exit (0);
    }

    init_mrt (default_trace);
    read_ascii (filename, IO);
    exit (0);
}


static void 
read_ascii (char *name, io_t *IO)
{

    FILE *fd;
    char line[MAXLINE], *ret;
    char data[MAXLINE];
    bgp_attr_t *attr;
    int announce_flag;
    LINKED_LIST *ll_ann, *ll_with;
    prefix_t *prefix;
    u_int size;
    u_char *cp;
    int line_num = 0;
    int have_data = 0;
    u_int as, value;
    time_t tstamp = 0;
    enum MRT_MSG_TYPES type = 0;
    enum MRT_MSG_BGP_TYPES bgp_type = 0;
    prefix_t *to = NULL;
    u_int AS = 0;

    if (!strcmp (name, "stdin"))
	fd = stdin;
    else
	fd = fopen (name, "r");

    if (fd == NULL) {
	perror ("Error opening input");
	exit (1);
    }

    announce_flag = 0;
    attr = bgp_new_attr (PROTO_BGP);
    ll_ann = LL_Create (LL_DestroyFunction, Deref_Prefix, 0);
    ll_with = LL_Create (LL_DestroyFunction, Deref_Prefix, 0);

    line_num++;
    if (fgets (line, MAXLINE - 1, fd) == NULL)
	return;
    while (line[0] == '\n') {
	line_num++;
	if (fgets (line, MAXLINE - 1, fd) == NULL)
	    return;
    }
    for (;;) {
	trace (NORM, MRT->trace, "%s", line);

	/* TIME */
	if (!strncmp (line, "TIME: ", 6)) {
	    struct tm tm;
	    if (strptime (line + 6, "%D %T", &tm)) {
		time_t now;
		time (&now);
		tm.tm_isdst = localtime (&now)->tm_isdst;
		tstamp = mktime (&tm);
	    }
	}

	/* TO: */
	if (!strncmp (line, "TO: ", 4)) {
	    if (sscanf (line + 4, "AS%d %s", &AS, data) >= 2) {
		if (to)
		    Deref_Prefix (to);
		to = ascii2prefix (0, data);
	    }
	}

	/* FROM: */
	if (!strncmp (line, "FROM: ", 6)) {
	    if (sscanf (line + 6, "%s AS%d", data, &AS) >= 2) {
		prefix = ascii2prefix (0, data);
		attr->gateway = add_gateway (prefix, AS, NULL);
		Deref_Prefix (prefix);
	    }
	}

	/* type BGP/UPDATE */
	if (sscanf (line, "TYPE: %s", data) >= 1) {
	    char *subtypestr;
	    char **cpp;
	    int subtype = 0;
	    int i;

	    if ((subtypestr = strchr (data, '/')) != NULL)
		*subtypestr++ = '\0';

	    cpp = S_MRT_MSG_TYPES;
	    for (i = 0; cpp[i]; i++) {
	        if (strcasecmp (cpp[i], data) == 0)
			break;
	    }
	    type = (enum MRT_MSG_TYPES) i;

	    if (subtypestr) {
		if ((cpp = S_MRT_MSG_SUBTYPES[type]) != NULL) {
		    for (i = 0; cpp[i]; i++) {
		        if (strcasecmp (cpp[i], subtypestr) == 0)
			    break;
		    }
		    subtype = i;
		}
	    }

	    switch (type) {
		case MSG_PROTOCOL_BGP:
#ifdef HAVE_IPV6
		case MSG_PROTOCOL_BGP4PLUS:
		case MSG_PROTOCOL_BGP4PLUS_01:
#endif /* HAVE_IPV6 */
		    bgp_type = (enum MRT_MSG_BGP_TYPES) subtype;

		    if (bgp_type == MSG_BGP_UPDATE) {
		        have_data++;
		    }
		    else {
		        trace (ERROR, MRT->trace, 
			       "*** ERROR Unknown message subtype %s"
			       " at line %d\n",
			       subtypestr, line_num);
		        skip (fd);
		    }
		    break;
		default:
		    trace (ERROR, MRT->trace, 
			   "*** ERROR Unknown message type %s at line %d\n",
			   data, line_num);
		    skip (fd);
		    break;
	    }
	}

	/* announce/withdraw */
	else if (!strncmp ("ANNOUNCE", line, 8)) {
	    announce_flag = 1;
	}
	else if (!strncmp (line, "WITHDRAW", 8)) {
	    announce_flag = 0;
	}

	/* origin */
	else if (sscanf (line, "ORIGIN: %s", data) >= 1) {

	    if (strcasecmp (data, "IGP") == 0) {
		attr->origin = 0;
		BGP4_BIT_SET (attr->attribs, PA4_TYPE_ORIGIN);
	    }
	    else if (strcasecmp (data, "EGP") == 0) {
		attr->origin = 1;
		BGP4_BIT_SET (attr->attribs, PA4_TYPE_ORIGIN);
	    }
	    else if (strcasecmp (data, "INCOMPLETE") == 0) {
		attr->origin = 2;
		BGP4_BIT_SET (attr->attribs, PA4_TYPE_ORIGIN);
	    }
	    else
		trace (ERROR, MRT->trace, "Unrecognized ORIGIN line %d\n", 
		       line_num);
	}

	/* aspath */
	else if (!strncmp (line, "ASPATH: ", 8)) {
	    if ((attr->aspath = aspth_from_string (line + 8)) == NULL)
		trace (ERROR, MRT->trace, "Unrecognized ASPATH line %d\n", 
		       line_num);
	    else
		BGP4_BIT_SET (attr->attribs, PA4_TYPE_ASPATH);
	}

	/* next hop */
	else if (sscanf (line, "NEXT_HOP: %s", data) >= 1) {
#ifdef HAVE_IPV6
	    if ((prefix = ascii2prefix (0, data)) == NULL)
		trace (ERROR, MRT->trace, "Unrecognized NEXT_HOP line %d\n", 
		       line_num);
	    else {
		if (prefix->family == AF_INET6) {

		    if (attr->nexthop && attr->nexthop->family == AF_INET) {
			attr->nexthop4 = attr->nexthop;		/* save */
			attr->nexthop = NULL;
		    }

		    if (ipv6_link_local_addr (prefix_toaddr6 (prefix))) {
		        attr->link_local = prefix;
		        if (attr->nexthop == NULL)
			    attr->nexthop = Ref_Prefix (prefix);
		    }
		    else {
			if (attr->nexthop)
			    Deref_Prefix (attr->nexthop);
		        attr->nexthop = prefix;
		    }
		}
		else {
		    if (attr->nexthop && attr->nexthop->family == AF_INET6) {
		            attr->nexthop4 = prefix;
		    }
		    else {
			if (attr->nexthop)
			    Deref_Prefix (attr->nexthop);
		        attr->nexthop = prefix;
		    }
		}
		BGP4_BIT_SET (attr->attribs, PA4_TYPE_NEXTHOP);
	    }
#else
	    if ((attr->nexthop = ascii2prefix (AF_INET, data)) == NULL)
		trace (ERROR, MRT->trace, "Unrecognized NEXT_HOP line %d\n", 
		       line_num);
	    else
		BGP4_BIT_SET (attr->attribs, PA4_TYPE_NEXTHOP);
#endif /* HAVE_IPV6 */
	}

	else if (sscanf (line, "MULTIEXIT: %d", &value) >= 1 ||
	         sscanf (line, "MULTI_EXIT_DISC: %d", &value) >= 1) {
	    attr->multiexit = value;
	    BGP4_BIT_SET (attr->attribs, PA4_TYPE_METRIC);
	}

	else if (sscanf (line, "LOCAL_PREF: %d", &value) >= 1) {
	    attr->local_pref = value;
	    BGP4_BIT_SET (attr->attribs, PA4_TYPE_LOCALPREF);
	}

	else if (strncmp (line, "ATOMIC_AGGREGATE", 16) == 0) {
	    BGP4_BIT_SET (attr->attribs, PA4_TYPE_ATOMICAGG);
	}

	else if (sscanf (line, "AGGREGATOR: AS%d %s", &as, data) >= 1) {
	    attr->aggregator.as = as;
	    if ((attr->aggregator.prefix = ascii2prefix (AF_INET, data)) 
		    == NULL)
		trace (ERROR, MRT->trace,
		       "Unrecognized AGGREGATOR line %d\n", line_num);
	    else
		BGP4_BIT_SET (attr->attribs, PA4_TYPE_AGGREGATOR);
	}

	else if (!strncmp (line, "COMMUNITY: ", 11)) {
	    if ((attr->community = community_from_string (line + 11)) == NULL)
		trace (ERROR, MRT->trace, "Unrecognized COMMUNITY line %d\n", 
		       line_num);
	    else
		BGP4_BIT_SET (attr->attribs, PA4_TYPE_COMMUNITY);
	}

	else if (sscanf (line, "DPA: AS%d %d", &as, &value) >= 2) {
	    attr->dpa.as = as;
	    attr->dpa.value = value;
	    BGP4_BIT_SET (attr->attribs, PA4_TYPE_DPA);
	}

	/* a prefix -- line begins with a space */
	else if (line[0] == ' ' || line[0] == '\t') {
	    line[strlen (line) - 1] = '\0';
	    if ((prefix = ascii2prefix (0, (line + 2))) == NULL) {
		trace (ERROR, MRT->trace, "Unrecognized prefix line %d\n", 
		       line_num);
	    }
	    else {
		if (announce_flag == 1) {
		    LL_Add (ll_ann, prefix);
		}
		else {
		    LL_Add (ll_with, prefix);
		}
	    }
	}

	/* try to get next line */
	line_num++;
	if ((ret = fgets (line, MAXLINE - 1, fd)) == NULL || line[0] == '\n') {
	    /* eof or newline */
	    if (have_data) {
		cp = bgp_create_update_msg (type, &size, attr, ll_ann, ll_with,
					    AS, to);
		io_write (IO, tstamp, type, bgp_type, size, (char *) cp);
		Delete (cp);
		have_data = 0;
		trace (NORM, MRT->trace, "***\n");
	    }
	    AS = 0;
	    if (to) {
		Deref_Prefix (to);
		to = NULL;
	    }
	    trace (NORM, MRT->trace, "\n");
	    bgp_deref_attr (attr);
	    if (ret == NULL) {
		LL_Destroy (ll_ann);
		LL_Destroy (ll_with);
		return;
	    }
	    while (line[0] == '\n') {
		line_num++;
		if (fgets (line, MAXLINE - 1, fd) == NULL)
		    return;
	    }
	    announce_flag = 0;
	    LL_Clear (ll_ann);
	    LL_Clear (ll_with);
	    attr = bgp_new_attr (PROTO_BGP);
	}
    }
}


static void 
skip (FILE * fd)
{
    char line[MAXLINE];

    while (fgets (line, MAXLINE - 1, fd) != NULL) {
	if (line[0] == '\n')
	    return;
    }
    return;
}


#ifndef HAVE_STRPTIME

/*
 * Please be careful because this function doesn't consult the format.
 * This only parses "MM/DD/YY HH:MM:SS" like 05/06/96 23:55:10.
 */

char *
strptime (char *buf, const char *format, struct tm *tm)
{

    static char *tm_zone = NULL;
    static long tm_gmtoff = 0;
    static int tm_isdst = 0;

    if (strlen (buf) < 17)
	return (NULL);

    if (tm_zone == NULL && tm_gmtoff == 0) {
	struct tm *tmp;
	time_t now;
	time (&now);
	tmp = localtime (&now);
	tm_isdst = tmp->tm_isdst;
	tm_gmtoff = tmp->tm_gmtoff;
	tm_zone = strdup (tmp->tm_zone);
    }
    buf[2] = buf[5] = buf[8] = buf[11] = buf[14] = '\0';
    tm->tm_mon = atoi (buf + 0) - 1;
    tm->tm_mday = atoi (buf + 3);
    tm->tm_year = atoi (buf + 6);
    tm->tm_hour = atoi (buf + 9);
    tm->tm_min = atoi (buf + 12);
    tm->tm_sec = atoi (buf + 15);
    tm->tm_gmtoff = tm_gmtoff;
    tm->tm_zone = tm_zone;

    return (buf + 17);
}

#endif /* HAVE_STRPTIME */
